import { TricksterInventory } from "./trickster"
import {ColumnInfo, ColumnName, Columns, ColumnSorter, LazyColumn} from "./columns"
import { PredefinedMenuItemKey } from "handsontable/plugins/contextMenu"
import Handsontable from "handsontable"
import { HotTableProps } from "@handsontable/react"



export interface InventoryTableOptions {
  columns: ColumnSet
  tags: ColumnSet

  accounts: string[]
}

export interface Mappable<T> {
  map<U>(callbackfn: (value: T, index: number, array: T[]) => U, thisArg?: any): U[];
}
export class ColumnSet implements  Mappable<ColumnInfo>{
  s: Set<ColumnName> = new Set()
  size: number;
  dirty = 0
  constructor(i?:Iterable<ColumnInfo | ColumnName>){
    if(i){
      for (const a of i) {
        if(Columns[LazyColumn(a)]){
          this.s.add(LazyColumn(a))
        }
      }
    }
    this.size = 0
    this.mark()
  }
  map<U>(callbackfn: (value: ColumnInfo, index: number, array: ColumnInfo[]) => U, thisArg?: any): U[] {
    return Array.from(this.values()).map(callbackfn, thisArg)
  }
  [Symbol.iterator](): IterableIterator<ColumnInfo>{
    return this.values()
  }
  [Symbol.toStringTag] = "ColumnSet";
  entries(): IterableIterator<[ColumnInfo, ColumnInfo]>{
    return Array.from(this.values()).map((x):[ColumnInfo,ColumnInfo]=>{return [x,x]}).values()
  }
  keys(): IterableIterator<ColumnInfo>{
    return this.values()
  }
  forEach(callbackfn: (value: ColumnInfo, value2: ColumnInfo, set: Set<ColumnInfo>) => void, thisArg?: any): void{
    Array.from(this.values()).forEach((v)=>{
      if(this.has(v)) {
        callbackfn(v, v, new Set(this.values()))
      }
    }, thisArg)
  }
  values(): IterableIterator<ColumnInfo>{
    return Array.from(this.s.values()).sort(ColumnSorter).map((a, b)=>{
      return Columns[a]
    }).values()
  }
  mark() {
    this.dirty= this.dirty+ 1
    this.size = this.s.size
  }
  add(value: ColumnInfo | ColumnName): this {
    this.mark()
    this.s.add(LazyColumn(value))
    return this
  }
  clear(): void {
    this.mark()
    this.s.clear()
  }
  delete(value: ColumnInfo | ColumnName): boolean{
    this.mark()
    return this.s.delete(LazyColumn(value))
  }
  has(value: ColumnInfo | ColumnName): boolean{
    return this.s.has(LazyColumn(value))
  }
}

export class InventoryTable {
  inv!:TricksterInventory
  o!: InventoryTableOptions

  constructor(inv:TricksterInventory, o:InventoryTableOptions) {
    this.setInv(inv)
    this.setOptions(o)
  }
  setOptions(o:InventoryTableOptions) {
    this.o = o
  }
  setInv(inv:TricksterInventory) {
    this.inv = inv
  }
  getTableColumnNames(): string[] {
    return this.o.columns.map(x=>x.displayName)
  }
  getTableColumnSettings(): Handsontable.ColumnSettings[] {
    return this.o.columns.map(x=>{
      let out:Handsontable.ColumnSettings = {
        renderer: x.renderer ? x.renderer : "text",
        filters: true,
        dropdownMenu: x.filtering ? DefaultDropdownItems() : false,
        readOnly: x.writable ? false : true,
        selectionMode: (x.writable ? "multiple" : 'single') as any,
      }
      if(x.options) {
        out.type = 'dropdown'
        if(typeof x.options == "function") {
        out.source = x.options(this.o.accounts)
        }else {
          out.source = x.options
        }
      }
      return out
    })
  }
  getTableRows():any[][] {
    return Object.values(this.inv.items)
    .filter((item):boolean=>{
      if(item.item_count <= 0) {
        return false
      }
      let found = true
      let hasAll = this.o.tags.has("All")
      if(this.o.tags.s.size > 0) {
        found = hasAll
        for(const tag of this.o.tags.values()) {
          if(tag.name =="All") {
            continue
          }
          if(tag.getter(item) === 1) {
              return !hasAll
          }
        }
      }
      return found
    })
    .map((item)=>{
      return this.o.columns.map(x=>{
        return x.getter(item)
      })
    })
  }
  BuildTable():TableRecipe {
    const s = DefaultSettings()
    const dat = this.getTableRows()
    return {
      data: dat,
      settings: {
        data: dat,
        colHeaders:this.getTableColumnNames(),
        columns:this.getTableColumnSettings(),
        ...s
      },
    }
  }
}

export interface TableRecipe {
  data: any[][]
  settings: HotTableProps
}

export const DefaultDropdownItems = ():PredefinedMenuItemKey[]=>['filter_by_condition' , 'filter_operators' ,'filter_by_condition2' , 'filter_by_value' , 'filter_action_bar']
export const DefaultSettings = ():HotTableProps=>{
  return {
    trimDropdown: true,
    filters: true,
    manualRowMove: false,
    manualColumnMove: false,
    allowInsertRow: false,
    allowInsertColumn: false,
    allowRemoveRow: false,
    allowRemoveColumn: false,
    allowHtml: true,
    disableVisualSelection: false,
    columnSorting: {
      indicator: true,
      headerAction: true,
    },
    hiddenColumns: {
      columns: [0],
    },
   // renderAllRows: true,
    viewportColumnRenderingOffset: 3,
    viewportRowRenderingOffset: 10,
 //   dropdownMenu: DefaultDropdownItems(),
    afterGetColHeader: (col, th)  => {
      if(!th.innerHTML.toLowerCase().includes("name")) {
        const ct = th.querySelector('.changeType')
        if(ct){
          ct.parentElement!.removeChild(ct)
        }
      }
    },
    beforeOnCellMouseDown(event:any, coords) {
      // Deselect the column after clicking on input.
      if (coords.row === -1 && event.target.nodeName === 'INPUT') {
        event.stopImmediatePropagation();
      }
    },
    className: 'htLeft',
    contextMenu: false,
    readOnlyCellClassName: "",
    licenseKey:"non-commercial-and-evaluation",
  }
}
